# purerouter-sdk/src/purerouter/resources/deployments.py
from __future__ import annotations
from typing import Any, Dict, Generator, AsyncGenerator

import httpx

from ..errors import APIError
from ..types import InvokeRequest, StreamEvent


class DeploymentsResource:
    def __init__(self, client: Any) -> None:
        self._client = client

    def _to_payload(self, request: InvokeRequest) -> Dict[str, Any]:
        if hasattr(request, "to_payload") and callable(getattr(request, "to_payload")):
            return request.to_payload()  # type: ignore[attr-defined]
        return {k: v for k, v in request.__dict__.items() if v is not None}

    def invoke(self, deployment_id: str, request: InvokeRequest) -> Dict[str, Any]:
        """
        Synchronous call without streaming.        
        """
        url = f"{self._client.base_url}/v1/deployments/{deployment_id}/invoke"
        payload = self._to_payload(request)
        with self._client._client() as c:  # pylint: disable=protected-access
            req = c.build_request("POST", url, json=payload)
            resp = self._client._retry_send(c, req)  # pylint: disable=protected-access
            if resp.status_code >= 400:
                raise APIError(resp.status_code, resp.text, body=resp.text)
            return resp.json()

    def stream(
        self,
        deployment_id: str,
        request: InvokeRequest,
    ) -> Generator[StreamEvent, None, None]:
        """
        Synchronous streaming call (SSE)..
        """
        url = f"{self._client.base_url}/v1/deployments/{deployment_id}/invoke"
        payload = self._to_payload(request)
        payload["stream"] = True

        headers = {"Accept": "text/event-stream"}
        with self._client._client() as c:  # pylint: disable=protected-access
            with c.stream("POST", url, json=payload, headers=headers) as resp:
                if resp.status_code >= 400:
                    body = resp.read().decode("utf-8", "ignore")
                    raise APIError(resp.status_code, body, body=body)
                yield from self._client._iter_stream(resp)  

    # Async
    async def ainvoke(self, deployment_id: str, request: InvokeRequest) -> Dict[str, Any]:
        """
        Asynchronous call without streaming.
        """
        url = f"{self._client.base_url}/v1/deployments/{deployment_id}/invoke"
        payload = self._to_payload(request)
        async with self._client._client() as c:  # pylint: disable=protected-access
            resp = await c.post(url, json=payload)
            if resp.status_code >= 400:
                raise APIError(resp.status_code, resp.text, body=await resp.aread())
            return resp.json()

    async def astream(
        self,
        deployment_id: str,
        request: InvokeRequest,
    ) -> AsyncGenerator[StreamEvent, None]:
        """
        Streaming asynchronous invocation (SSE).
        """
        url = f"{self._client.base_url}/v1/deployments/{deployment_id}/invoke"
        payload = self._to_payload(request)
        payload["stream"] = True

        headers = {"Accept": "text/event-stream"}
        async with self._client._client() as c:  
            async with c.stream("POST", url, json=payload, headers=headers) as resp:
                if resp.status_code >= 400:
                    body = await resp.aread()
                    raise APIError(resp.status_code, body.decode("utf-8", "ignore"), body=body)
                async for ev in self._client._aiter_stream(resp):  
                    yield ev



# purerouter-sdk/src/purerouter/client.py
from __future__ import annotations
import time
from typing import Any, Dict, Iterable, Optional

import httpx

from .errors import APIError
from .types import Json, StreamEvent
from .resources.router import RouterResource
from .resources.deployments import DeploymentsResource

DEFAULT_BASE_URL = "https://api.purerouter-api.com"


class _Base:
    def __init__(
        self,
        *,
        router_key: Optional[str] = None,
        base_url: str = DEFAULT_BASE_URL,
        timeout: float = 30.0,
        headers: Optional[Dict[str, str]] = None,
    ) -> None:
        self.base_url = base_url.rstrip("/")
        self.timeout = timeout
        self.default_headers = {
            "User-Agent": "purerouter-python/0.1",
        }
        if headers:
            self.default_headers.update(headers)
        if router_key:
            self.default_headers.setdefault("x-router-key", router_key)

    def _retry_send(self, client: httpx.Client, req: httpx.Request) -> httpx.Response:
        backoff = 0.25
        for attempt in range(5):
            try:
                resp = client.send(req)
                if resp.status_code >= 500:
                    raise APIError(resp.status_code, "Server error", body=resp.text)
                return resp
            except (httpx.TransportError, APIError):
                if attempt == 4:
                    raise
                time.sleep(backoff)
                backoff *= 2
        raise RuntimeError("unreachable")

    @staticmethod
    def _iter_stream(resp: httpx.Response) -> Iterable[StreamEvent]:
        ct = (resp.headers.get("content-type") or "").lower()
        if "text/event-stream" in ct:
            for line in resp.iter_lines():
                if not line:
                    continue
                if isinstance(line, (bytes, bytearray)):
                    line = line.decode("utf-8", "ignore")
                if line.startswith(":"):
                    continue
                if line.lower().startswith("data:"):
                    line = line[5:].lstrip()
                yield StreamEvent(data=line)
        else:
            for chunk in resp.iter_bytes():
                if not chunk:
                    continue
                if isinstance(chunk, (bytes, bytearray)):
                    yield StreamEvent(data=chunk.decode("utf-8", "ignore"))
                else:
                    yield StreamEvent(data=str(chunk))

    @staticmethod
    async def _aiter_stream(resp: httpx.Response):
        ct = (resp.headers.get("content-type") or "").lower()
        if "text/event-stream" in ct:
            async for line in resp.aiter_lines():
                if not line:
                    continue
                if line.startswith(":"):
                    continue
                if line.lower().startswith("data:"):
                    line = line[5:].lstrip()
                yield StreamEvent(data=line)
        else:
            async for chunk in resp.aiter_bytes():
                if not chunk:
                    continue
                if isinstance(chunk, (bytes, bytearray)):
                    yield StreamEvent(data=chunk.decode("utf-8", "ignore"))
                else:
                    yield StreamEvent(data=str(chunk))


class PureRouter(_Base):

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.router = RouterResource(self)
        self.deployments = DeploymentsResource(self)

    def _client(self) -> httpx.Client:
        return httpx.Client(timeout=self.timeout, headers=self.default_headers)

    def health(self) -> Json:
        url = f"{self.base_url}/v1/health"
        with self._client() as c:
            r = self._retry_send(c, c.build_request("GET", url))
            if r.status_code >= 400:
                raise APIError(r.status_code, r.text, body=r.text)
            return r.json()


class AsyncPureRouter(_Base):

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.router = RouterResource(self)
        self.deployments = DeploymentsResource(self)

    def _client(self) -> httpx.AsyncClient:
        return httpx.AsyncClient(timeout=self.timeout, headers=self.default_headers)

    async def health(self) -> Json:
        url = f"{self.base_url}/v1/health"
        async with self._client() as c:
            r = await c.get(url)
            if r.status_code >= 400:
                raise APIError(r.status_code, r.text, body=r.text)
            return r.json()



# purerouter-sdk/src/purerouter/types.py
from __future__ import annotations
from dataclasses import dataclass
from typing import Any, Dict, Iterable, Optional

Json = Dict[str, Any]


@dataclass
class RequestOptions:
    timeout: Optional[float] = 30.0
    headers: Optional[Dict[str, str]] = None


@dataclass
class InferRequest:
    prompt: str
    profile: Optional[str] = None
    stream: Optional[bool] = None

@dataclass
class InferResponse:
    output_text: Optional[str] = None
    provider: Optional[str] = None
    model_id: Optional[str] = None
    task: Optional[str] = None
    profile: Optional[str] = None
    latency_ms: Optional[int] = None
    usage: Optional[Json] = None
    raw: Optional[Any] = None
    cost_usd: Optional[float] = None


# ---------- Deployments (invoke) ----------
@dataclass
class InvokeRequest:

    messages: Optional[Iterable[Json]] = None

    input: Optional[Json] = None
    inputs: Optional[Json] = None

    prompt: Optional[str] = None
    max_tokens: Optional[int] = None
    temperature: Optional[float] = None

    parameters: Optional[Json] = None
    stream: Optional[bool] = None


# ---------- Streaming ----------
@dataclass
class StreamEvent:
    data: str

__all__ = [
    "Json",
    "RequestOptions",
    "InferRequest",
    "InferResponse",
    "InvokeRequest",
    "StreamEvent",
]



# purerouter-sdk/src/purerouter/resources/router.py
from __future__ import annotations
from typing import Any, Generator

import httpx
from dataclasses import fields as dc_fields

from ..errors import APIError
from ..types import InferRequest, InferResponse, StreamEvent


class RouterResource:
    def __init__(self, client: Any) -> None:
        self._client = client

    def infer(self, request: InferRequest) -> InferResponse:
        url = f"{self._client.base_url}/v1/infer"
        payload = {k: v for k, v in request.__dict__.items() if v is not None}
        with self._client._client() as c:  # pylint: disable=protected-access
            req = c.build_request("POST", url, json=payload)
            resp = self._client._retry_send(c, req)  # pylint: disable=protected-access
            if resp.status_code >= 400:
                body = resp.text
                raise APIError(resp.status_code, body, body=body)
            data = resp.json()
            allowed = {f.name for f in dc_fields(InferResponse)}
            filtered = {k: v for k, v in data.items() if k in allowed}
            return InferResponse(**filtered)

    def stream(self, request: InferRequest) -> Generator[StreamEvent, None, None]:
        url = f"{self._client.base_url}/v1/infer"
        payload = {k: v for k, v in request.__dict__.items() if v is not None}
        payload["stream"] = True
        with self._client._client() as c:  
            with c.stream("POST", url, json=payload) as resp:
                if resp.status_code >= 400:
                    body = resp.read().decode("utf-8", "ignore")
                    raise APIError(resp.status_code, body, body=body)
                yield from self._client._iter_stream(resp)  # pylint: disable=protected-access


    async def ainfer(self, request: InferRequest) -> InferResponse:
        url = f"{self._client.base_url}/v1/infer"
        payload = {k: v for k, v in request.__dict__.items() if v is not None}
        async with self._client._client() as c:  # pylint: disable=protected-access
            resp = await c.post(url, json=payload)
            if resp.status_code >= 400:
                raise APIError(resp.status_code, resp.text, body=await resp.aread())
            data = resp.json()
            allowed = {f.name for f in dc_fields(InferResponse)}
            filtered = {k: v for k, v in data.items() if k in allowed}
            return InferResponse(**filtered)

    async def astream(self, request: InferRequest):
        url = f"{self._client.base_url}/v1/infer"
        payload = {k: v for k, v in request.__dict__.items() if v is not None}
        payload["stream"] = True
        async with self._client._client() as c:  
            async with c.stream("POST", url, json=payload) as resp:
                if resp.status_code >= 400:
                    body = await resp.aread()
                    raise APIError(resp.status_code, body.decode("utf-8", "ignore"), body=body)
                async for ev in self._client._aiter_stream(resp):  # pylint: disable=protected-access
                    yield ev



# purerouter-sdk/src/purerouter/errors.py
from typing import Optional

class APIError(Exception):
    def __init__(self, status_code: int, message: str, *, body: Optional[object] = None):
        super().__init__(f"HTTP {status_code}: {message}")
        self.status_code = status_code
        self.body = body